import numpy as np


nan = np.nan
r"NaN (Not a Number)"

inf = np.inf
r":math:`\infty` (infinite)"

pi = np.pi
r":math:`\pi`"

e = np.e
r":math:`e`"

euler_gamma = np.euler_gamma
r"Euler's :math:`\gamma`"
